-- $Id$
-- ToME - Tales of Maj'Eyal
-- Copyright © 2020 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham ("Zizzo")
-- dsb-tome@killerbunnies.org

local _M = loadPrevious(...)

local UI = require "engine.ui.Base"
local Shader = require "engine.Shader"
local Dialog = require "engine.ui.Dialog"

local frames_colors = {
	ok = {0.3, 0.6, 0.3},
	sustain = {0.6, 0.6, 0},
	cooldown = {0.6, 0, 0},
	disabled = {0.65, 0.65, 0.65},
}

local super_init = _M.init
function _M:init()
  super_init(self)

  -- Spectacularly Evil But Awesome Hack(TM):  Since apparently none of the
  -- text-displaying engine code can display a '#' character without
  -- thinking it's part of a TString highlighting sequence, we make an
  -- Entity with a '#' in it to display in its place, using deep wizardry
  -- to force the width of the entity to the width of the character.
  local Entity = require 'engine.Entity'
  local Base = require 'engine.ui.Base'
  self.pound = Entity.new {
    display = '#',
    color = colors.WHITE,
    force_w = Base.font_mono:size('M'),
    getEntityFinalSurface = function(e_self, tiles, w, h)
      return require('engine.Entity').getEntityFinalSurface(e_self, tiles, e_self.force_w, h)
    end,
  }
end

-- Replace this method with a copy that wraps the effect's long_desc() in a
-- pcall() to catch and hilight errors.
function _M:handleEffect(player, eff_id, e, p, x, y, hs, bx, by, is_first, scale, allow_remove)
  local shader = Shader.default.textoutline and Shader.default.textoutline.shad

  local dur = p.dur + 1
  local charges = e.charges and tostring(e.charges(player, p)) or "0"

  if not self.tbuff[eff_id..":"..dur..":"..charges] then
    local name = e.desc
    local desc = nil
    local eff_subtype = table.concat(table.keys(e.subtype), "/")
    if e.display_desc then name = e.display_desc(self, p) end
    -- Our main change:  wrap the call to e.long_desc() in a Lua pcall() to
    -- catch any errors.
    local ok_long_desc, long_desc = pcall(e.long_desc, player, p)
    if not ok_long_desc then
      -- Spectacularly Evil But Awesome Hack(TM), redux:  "Escape" '#'
      -- characters in the error message with our entity above.
      local pound_txt = self.pound:getDisplayString()
      long_desc = long_desc:gsub('#', pound_txt)
    end
    if p.save_string and p.amount_decreased and p.maximum and p.total_dur then
      desc = ("#{bold}##GOLD#%s\n(%s: %s)#WHITE##{normal}#\n"):format(name, e.type, eff_subtype)..long_desc.." "..("%s reduced the duration of this effect by %d turns, from %d to %d."):format(p.save_string, p.amount_decreased, p.maximum, p.total_dur)
    else
      desc = ("#{bold}##GOLD#%s\n(%s: %s)#WHITE##{normal}#\n"):format(name, e.type, eff_subtype)..long_desc
    end
    if allow_remove then desc = desc.."\n---\nRight click to cancel early." end

    local txt = nil
    local txt2 = nil
    if e.decrease > 0 then
      local font = e.charges and self.buff_font_small or self.buff_font
      dur = tostring(dur)
      txt = font:draw(dur, 40, colors.WHITE.r, colors.WHITE.g, colors.WHITE.b, true)[1]
      txt.fw, txt.fh = font:size(dur)
    end
    if e.charges then
      local font = e.decrease > 0 and self.buff_font_smallmed or self.buff_font

      txt2 = font:draw(charges, 40, colors.WHITE.r, colors.WHITE.g, colors.WHITE.b, true)[1]
      txt2.fw, txt2.fh = font:size(charges)
      dur = dur..":"..charges
    end
    local icon = e.status ~= "detrimental" and frames_colors.ok or frames_colors.cooldown
    -- Additional small change:  If the effect's long_desc() threw an
    -- error, mark the effect's icon with a blue border (we can't use red,
    -- because the UI already uses that for detrimental effects).
    if not ok_long_desc then icon = { 0, 1.0, 1.0 } end

    local desc_fct = function(button, mx, my, xrel, yrel, bx, by, event)
      if is_first then
	if event == "out" then self.mhandle.buffs = nil return
	else self.mhandle.buffs = true end

	-- Move handle
	if not self.locked and bx >= self.mhandle_pos.buffs.x and bx <= self.mhandle_pos.buffs.x + move_handle[6] and by >= self.mhandle_pos.buffs.y and by <= self.mhandle_pos.buffs.y + move_handle[7] then self:uiMoveResize("buffs", button, mx, my, xrel, yrel, bx, by, event) end
      end
      if config.settings.cheat and event == "button" and core.key.modState("shift") then
	if button == "left" then
	  p.dur = p.dur + 1
	elseif button == "right" then
	  p.dur = p.dur - 1
	end
      elseif allow_remove and event == "button" and button == "right" then
	Dialog:yesnoPopup(name, "Really cancel "..name.."?", function(ret)
	  if ret then
	    player:removeEffect(eff_id)
	  end
	end)
      end
      game.tooltip_x, game.tooltip_y = 1, 1; game:tooltipDisplayAtMap(game.w, game.h, desc)
    end

    local flash = 0

    if p.__set_time and p.__set_time >= self.now - 500 then
      flash = 5
    end

    self.tbuff[eff_id..":"..dur..":"..charges] = {eff_id, "tbuff"..eff_id, function(x, y)
      core.display.drawQuad(x+4, y+4, 32, 32, 0, 0, 0, 255)
      e.display_entity:toScreen(self.hotkeys_display_icons.tiles, x+4, y+4, 32, 32)
      if e.get_fractional_percent then
	core.display.drawQuadPart(x +4 , y + 4, 32, 32, e.get_fractional_percent(self, p), 128, 128, 128, 200)
      end
      UI:drawFrame(self.buffs_base, x, y, icon[1], icon[2], icon[3], 1)

      if txt and not txt2 then
	if shader then
	  shader:use(true)
	  shader:uniOutlineSize(1, 1)
	  shader:uniTextSize(txt._tex_w, txt._tex_h)
	else
	  txt._tex:toScreenFull(x+4+2 + (32 - txt.fw)/2, y+4+2 + (32 - txt.fh)/2, txt.w, txt.h, txt._tex_w, txt._tex_h, 0, 0, 0, 0.7)
	end
	txt._tex:toScreenFull(x+4 + (32 - txt.fw)/2, y+4 + (32 - txt.fh)/2, txt.w, txt.h, txt._tex_w, txt._tex_h)
      elseif not txt and txt2 then
	if shader then
	  shader:use(true)
	  shader:uniOutlineSize(1, 1)
	  shader:uniTextSize(txt2._tex_w, txt2._tex_h)
	else
	  txt2._tex:toScreenFull(x+4+2, y+4+2 + (32 - txt2.fh)/2+5, txt2.w, txt2.h, txt2._tex_w, txt2._tex_h, 0, 0, 0, 0.7)
	end
	txt2._tex:toScreenFull(x+4, y+4 + (32 - txt2.fh)/2+5, txt2.w, txt2.h, txt2._tex_w, txt2._tex_h, 0, 1, 0, 1)
      elseif txt and txt2 then
	if shader then
	  shader:use(true)
	  shader:uniOutlineSize(1, 1)
	  shader:uniTextSize(txt._tex_w, txt._tex_h)
	else
	  txt._tex:toScreenFull(x+4+2 + (32 - txt.fw), y+4+2 + (32 - txt.fh)/2-5, txt.w, txt.h, txt._tex_w, txt._tex_h, 0, 0, 0, 0.7)
	end
	txt._tex:toScreenFull(x+4 + (32 - txt.fw), y+4 + (32 - txt.fh)/2-5, txt.w, txt.h, txt._tex_w, txt._tex_h)

	if shader then
	  shader:uniOutlineSize(1, 1)
	  shader:uniTextSize(txt2._tex_w, txt2._tex_h)
	else
	  txt2._tex:toScreenFull(x+4+1, y+4+6 + (32 - txt2.fh)/2+5, txt2.w, txt2.h, txt2._tex_w, txt2._tex_h, 0, 0, 0, 0.7)
	end
	txt2._tex:toScreenFull(x+3, y+8 + (32 - txt2.fh)/2+5, txt2.w, txt2.h, txt2._tex_w, txt2._tex_h, 0, 1, 0, 1)
      end

      if shader and (txt or txt2) then shader:use(false) end

      if flash > 0 then
	if e.status ~= "detrimental" then core.display.drawQuad(x+4, y+4, 32, 32, 0, 255, 0, 170 - flash * 30)
	else core.display.drawQuad(x+4, y+4, 32, 32, 255, 0, 0, 170 - flash * 30)
	end
	flash = flash - 1
      end
    end, desc_fct}
  end

  if not game.mouse:updateZone("tbuff"..eff_id, bx+x*scale, by+y*scale, hs, hs, self.tbuff[eff_id..":"..dur..":"..charges][4], scale) then
    game.mouse:unregisterZone("tbuff"..eff_id)
    game.mouse:registerZone(bx+x*scale, by+y*scale, hs, hs, self.tbuff[eff_id..":"..dur..":"..charges][4], nil, "tbuff"..eff_id, true, scale)
  end

  self.tbuff[eff_id..":"..dur..":"..charges][3](x, y)
end

return _M
